(*
defCFG - 
This module contains the definition of the Control Flow Graph datatype and the operations to work on it.
Authors: Pedro de Carvalho Gomes <pedrodcg@kth.se>
         Attilio Picoco <picoco@kth.se>
Date: 2013-02-02
*)

open Javalib_pack
open JCode
open Javalib
open JBasics
open JProgram
open JControlFlow
open JBir

(**
* Definition of the structure for the ControlFlowGraph
*)

type cfgFormat =
   | KTH
   | Normal

(**
* Mapping of the JBir.instr
*)
type edgeBirInstr =
   | Nop_Edge
   | AffectVar_Edge
   | AffectArray_Edge
   | AffectField_Edge
   | AffectStaticField_Edge
   | Goto_Edge
   | Ifd_Edge
   | Throw_Edge
   | Return_Edge
   | New_Edge
   | NewArray_Edge
   | InvokeStatic_Edge
   | InvokeVirtual_Edge
   | InvokeNonVirtual_Edge
   | MonitorEnter_Edge
   | MonitorExit_Edge
   | MayInit_Edge
   | Check_Edge

(**
* @param e edgeBirInstr BIR instruction
* @return int integer mapped
*
* It maps a BIR instruction edge to a number, for storing/loading purposes
*)
(* map_edge2int and map_int2edge should be modified coherently, if necessary *)
let map_edge2int e =
	match e with
	| Nop_Edge -> 1
	| AffectVar_Edge -> 2
	| AffectArray_Edge -> 3
	| AffectField_Edge -> 4
	| AffectStaticField_Edge -> 5
	| Goto_Edge -> 6
	| Ifd_Edge -> 7
	| Throw_Edge -> 8
	| Return_Edge -> 9
	| New_Edge -> 10
	| NewArray_Edge -> 11
	| InvokeStatic_Edge -> 12
	| InvokeVirtual_Edge -> 13
	| InvokeNonVirtual_Edge -> 14
	| MonitorEnter_Edge -> 15
	| MonitorExit_Edge -> 16
	| MayInit_Edge -> 17
	| Check_Edge -> 18

(**
* @param n int integer number
* @return edgeBirInstr BIR instruction mapped
*
* It maps a number to a BIR instruction edge, for storing/loading purposes
*)
(* map_edge2int and map_int2edge should be modified coherently, if necessary *)
let map_int2edge n =
	match n with
	| 1 -> Nop_Edge
	| 2 -> AffectVar_Edge
	| 3 -> AffectArray_Edge
	| 4 -> AffectField_Edge
	| 5 -> AffectStaticField_Edge
	| 6 -> Goto_Edge
	| 7 -> Ifd_Edge
	| 8 -> Throw_Edge
	| 9 -> Return_Edge
	| 10 -> New_Edge
	| 11 -> NewArray_Edge
	| 12 -> InvokeStatic_Edge
	| 13 -> InvokeVirtual_Edge
	| 14 -> InvokeNonVirtual_Edge
	| 15 -> MonitorEnter_Edge
	| 16 -> MonitorExit_Edge
	| 17 -> MayInit_Edge
	| 18 -> Check_Edge
	| _ -> failwith ("Wrong BIR instruction edge mapping. Leaving")

let map_bir2edge op =
	match op with
	| JBir.Nop -> Nop_Edge
	| JBir.AffectVar _ -> AffectVar_Edge
	| JBir.AffectArray _ -> AffectArray_Edge
	| JBir.AffectField _ -> AffectField_Edge
	| JBir.AffectStaticField _ -> AffectStaticField_Edge
	| JBir.Goto _ -> Goto_Edge
	| JBir.Ifd _ -> Ifd_Edge
	| JBir.Throw _ -> Throw_Edge
	| JBir.Return _ -> Return_Edge
	| JBir.New _ -> New_Edge
	| JBir.NewArray _ -> NewArray_Edge
	| JBir.InvokeStatic _ -> InvokeStatic_Edge
	| JBir.InvokeVirtual _ -> InvokeVirtual_Edge
	| JBir.InvokeNonVirtual _ -> InvokeNonVirtual_Edge
	| JBir.MonitorEnter _ -> MonitorEnter_Edge
	| JBir.MonitorExit _ -> MonitorExit_Edge
	| JBir.MayInit _ -> MayInit_Edge
	| JBir.Check _ -> Check_Edge

(** a node can take as labels: ret = return, ent = entry, exp = exception *)
type t_node_label = string

(** an edge can take class names as labels *)
type t_edge_label = string

(** Nodes in a CFG - contains the information about
 * methods's full signature, the index in the instruction array,
 * set of atomic propositions assigned to it and
 * a pair for unique identification from same nodes in same pc but with different semantics
 * *)
type t_cfg_node = {
   cmsig: class_method_signature;
   pc: int;
   mutable nlabels: t_node_label list;
   (* Added excp and ret to differentiate them from the same node in same pc *)
   excp: bool;
   ret: bool;
}


(** Defines the labeling function *)
type t_cfg_trans = {
(*   inst: JBir.instr; *)
	inst: edgeBirInstr;
   mutable elabels: t_edge_label list;
}

(** Defines the edges as triples *)
type t_cfg_edge = (t_cfg_node * t_cfg_trans * t_cfg_node)

(**
* @param x Triple (a,_,_)
* @return first object a
*
* Returns the first element of a triple *)
let get_first x = match x with (a,_,_) -> a;;
(**
* @param x Triple (_,b,_)
* @return second object b
*
* Returns the second element of a triple *)
let get_second x = match x with (_,b,_) -> b;;
(**
* @param x Triple (_,_,c)
* @return third object c
*
* Returns the third element of a triple *)
let get_third x = match x with (_,_,c) -> c;;

(** Define labels for entry and return nodes *)
let ret_lbl = "ret"
let ent_lbl = "entry"

(**
* @param n t_cfg_node
* @return string
*
* Generates a unique string *)
(* version with string *)
let get_node_id n = 
	let (cn,ms) = JBasics.cms_split n.cmsig
	in
	"c" ^ (string_of_int (JBasics.cn_hash cn)) ^ "m" ^ (string_of_int (JBasics.ms_hash ms)) ^ "p" ^ (string_of_int n.pc) ^ (if n.excp then ("-" ^ String.lowercase (List.hd (n.nlabels))) else "") ^ (if n.ret then "r" else "")
;;
(* version with hash number
let get_node_id n = 
	let (cn,ms) = JBasics.cms_split n.cmsig
	in
	"c" ^ (string_of_int (JBasics.cn_hash cn)) ^ "m" ^ (string_of_int (JBasics.ms_hash ms)) ^ "p" ^ (string_of_int n.pc) ^ (if n.excp then ("-" ^ string_of_int (Hashtbl.hash n.nlabels)) else "") ^ (if n.ret then "r" else "")
 *)

(**
* @param e t_cfg_edge
* @return string
*
* Generates a unique string *)
let get_edge_id e =
	"e" ^ (string_of_int (map_edge2int e.inst)) ^ " - labels: " ^ (List.fold_left (fun s l -> (s ^ " " ^ l)) "" e.elabels)


(** Define the CFG as a set of triples *)
module CFGSet = Set.Make (struct
   type t = t_cfg_edge
   let compare = compare
(*	let compare = (fun e1 e2 -> 
			(*
			let i1 = (Hashtbl.hash e1)
			and i2 = (Hashtbl.hash e2)
			in
			let c = Pervasives.compare i1 i2
			in
			print_string ("N1_e1: " ^ (get_node_id (get_first e1)) ^ "\n");
			print_string ("E_e1: " ^ (get_edge_id (get_second e1)) ^ "\n");
			print_string ("N2_e1: " ^ (get_node_id (get_third e1)) ^ "\n");
			print_string ("N1_e2: " ^ (get_node_id (get_first e2)) ^ "\n");
			print_string ("E_e2: " ^ (get_edge_id (get_second e2)) ^ "\n");
			print_string ("N2_e2: " ^ (get_node_id (get_third e2)) ^ "\n");
			print_string ("I1: " ^ (string_of_int i1) ^ " - I2: " ^ (string_of_int i2) ^ " - CMP: " ^ (string_of_int c) ^ "\n");
			c *)
			Pervasives.compare (Hashtbl.hash e1) (Hashtbl.hash e2)
			)*)
(*   let compare = (fun x y -> compare_nodes (get_first x) (get_first y)) *)
end)

(**
* @param tag_ret bool
* @param tag_excp bool
* @param tag_label t_node_label List
* @param pp JControlFlow.PP_Bir.t
* @param node_pc int
* @return t_cfg_node
*
* Function that creates nodes *)
let make_cfg_node ?(tag_ret=false) ?(tag_excp=false) ?(tag_label=[]) pp node_pc =
   {
      cmsig = (JControlFlow.PP_Bir.get_meth pp).cm_class_method_signature;
      pc = node_pc;
      nlabels = tag_label;
      excp = tag_excp;
      ret = tag_ret;
   }

(**
* @param tag_ret bool
* @param tag_excp bool
* @param tag_label t_node_label List
* @param pp JControlFlow.PP_Bir.t
* @param curr_meth_cms JBasics.class_method_signature
* @return t_cfg_node
*
* Function that creates nodes *)
let intra_make_cfg_node ?(tag_ret=false) ?(tag_excp=false) ?(tag_label=[]) p curr_meth_cms =
 { 
    cmsig = curr_meth_cms ;
    pc = (JControlFlow.PP_Bir.get_pc p);
    (* Add entry label only if PC is zero *)
    nlabels = if ( ((JControlFlow.PP_Bir.get_pc p) = 0) && (not tag_ret) && (not tag_excp)) then ([ent_lbl]@tag_label) else tag_label;
    excp = tag_excp; 
    ret = tag_ret;
 }

(**
* @param cms JBasics.class_method_signature
* @param pc int
* @param nl t_node_label List
* @param excp bool
* @param ret bool
* @return t_cfg_node
*
* Function that creates nodes *)
let make_cfg_node_pure cms pc nl excp ret =
{
	cmsig = cms;
	pc = pc;
	nlabels = nl;
	excp = excp;
	ret = ret;
}

(**
* @param op JBir.instr
* @param ls t_edge_label list
* @return t_cfg_edge
*
* Function that creates edges *)
let make_cfg_edge op ls =
	{ inst = (map_bir2edge op) ; elabels = ls }

(**
* @param op edgeBirInstr
* @param ls t_edge_label list
* @return t_cfg_edge
*
* Function that creates edges *)
let make_cfg_edge_pure op ls = { inst = op; elabels = ls }

(**
* @param source_node t_cfg_node
* @param edge t_cfg_trans
* @param target_list t_cfg_node list
* @param cfg CFGSet.t
* @return CFGSet.t
*
* Add edges from source_node to all target_nodes in target_list *)
let rec add_cfg_edge source_node edge target_list cfg =
   match target_list with
      | h::t-> CFGSet.add (source_node, edge, h) (add_cfg_edge source_node edge t cfg)
      | []-> cfg

(**
* @return t_cfg_node
*
* Function that creates a fake node
*)
let make_fake_node = 
{
	cmsig = (JBasics.make_cms JBasics.java_lang_object JBasics.clinit_signature);
	pc = -1;
	nlabels = [];
	excp = false;
	ret = false;
}

(**
* @return t_cfg_edge
*
* Function that creates a fake edge
*)
let make_fake_edge = 
{
	inst = Nop_Edge;
	elabels = [];
}
