(*
openEnv - 
This module contains the definition of the Open Environment datatype and the operations to work on it.
It is used in order to perform the modular extraction of CFG from Java bytecode within the tool 'fgextractor'.
It is also related to the module jMCA.
Authors: Pedro de Carvalho Gomes <pedrodcg@kth.se>
         Attilio Picoco <picoco@kth.se>
Version: 1.0
Date: 2012-06-08
*)

open Javalib_pack
open JCode
open Javalib
open JBasics
open JProgram


(**
* INTERFACE ENVIRONMENT Data structure
*)
type interface_env =
{
	ext_interfaces: JBasics.class_name list; (* Interfaces extended *)
	methods: JBasics.class_method_signature list (* Methods provided *)
}

(**
* CLASS ENVIRONMENT Data structure
*)
type class_env =
{
	super: JBasics.class_name option; (* Class name of the superclass - option as JBasics.java_lang_object does not have a superclass *)
	impl_interfaces: JBasics.class_name list; (* Interfaces implemented *)
	fields: JBasics.field_signature list (* Fields of the class - it suffices, no need for class_field_signature *)
}

(**
* METHOD ENVIRONMENT Data structure
*)
type method_env =
{
	(* If the method is not provided, it is faked as concrete with empty body - that is, at least a return instruction -> its code is marked as NONE *)
	code: int option; (* Hash value of the instruction array *)
	handlers: JBasics.class_name list; (* Exceptions NOT propagated - it can be an empty list *)
	req_meths: JBasics.class_method_signature list (* Client methods required -> used only for missing methods AND not in the check for refinement *)
}

(**
* OPEN ENVIRONMENT Data structure
*)
type open_env =
{
	interface_map: interface_env ClassMap.t; (* Partial mapping of interface names *)
	class_map: class_env ClassMap.t; (* Partial mapping of class names *)
	method_map: method_env ClassMethodMap.t (* Partial mapping of method references *)
}

(**
* @param im => interface_map 
* @param cm => class_map
* @param mm => method_map
* @return open_env
*)
val create_open_env : (interface_env ClassMap.t) -> (class_env ClassMap.t) -> (method_env ClassMethodMap.t) -> open_env

(**
* @param ei => extended_interfaces
* @param m => provided methods
* @return interface_env
*)
val create_int_env : (JBasics.class_name list) -> (JBasics.class_method_signature list) -> interface_env

(**
* @param s => super class 
* @param ii => implemented interfaces
* @param f => fields
* @return class_env
*)
val create_cl_env : (JBasics.class_name option) -> (JBasics.class_name list) -> (JBasics.field_signature list) -> class_env

(**
* @param c => hash value of the code
* @param h => handlers (exceptions not propagated)
* @return method_env
*)
val create_met_env : (int option) -> (JBasics.class_name list) -> (JBasics.class_method_signature list) -> method_env

(**
* @param old_oe => old Open Environment
* @param new_oe => new Open Environment
* @param all_meths_excp_map => JBasics.class_name list Javalib_pack.JBasics.ClassMethodMap.t Map of propagated exceptions for each cms
* @return bool
*
* This function checks if the new Open Environment is a refinement of the old one, following the rules given in the paper. 
*)
val is_a_refinement : open_env -> open_env -> (JBasics.class_name list) Javalib_pack.JBasics.ClassMethodMap.t -> bool

(**
* @param cms => class_method_signature to check if provided or not
* @param oe => open_env
* @return bool
*
* This checks if the method is provided or not
*)
val is_provided : JBasics.class_method_signature -> open_env -> bool

(**
* @return open_env
*
* This creates an empty OpenEnv
*)
val empty_open_env : open_env

(*
(**
* Name of the annotation to check against a method
*)
val annotation_cn : string
val annotation_elem_name_hndlr : string
val annotation_elem_name_rqmts : string
*)

(**
* @param oe => open_env
* @param cms => JBasics.class_method_signature
* @return JBasics.class_name list
*)
val get_handlers : open_env -> JBasics.class_method_signature -> (JBasics.class_name list)

(**
* Getting the list of children exceptions given a parent node + the parent. The list is made of class_names. It checks if the name is already in the list.
*)
val get_parent_children_excp : (Javalib_pack.JCode.jcode JProgram.program) -> (JBasics.class_name list) -> (JBasics.class_name) -> (JBasics.class_name list)

(**
* Getting the list of class_method_signature referenced in req_meths annotation
*)
val get_cms_from_rm_ann : (Javalib.method_annotations) -> (JBasics.class_method_signature list)

(**
* @param parsed_prog => 'a JProgram.program
* @param ref_classes => JBasics.class_name list
* @param ref_meths => JBasics.class_method_signature list
* @return OpenEnv.open_env
*
* It creates an Open Environment structure for the parsed program - only classes and methods of the client
*)
val get_open_env : Javalib_pack.JCode.jcode JProgram.program -> JBasics.class_name list -> JBasics.class_method_signature list -> open_env

(**
* @param old_oe => OpenEnv.open_env
* @param new_oe => OpenEnv.open_env
* @param cms => JBasics.class_method_signature
* @return bool
*
* It checks if a method has changed its code => either newly provided or removed or replaced
*)
val has_changed : open_env -> open_env -> JBasics.class_method_signature -> bool

(**
*@param oe => OpenEnv.open_env
*@return bool
*
* It checks if an Open Environment is properly defined or if it's empty
*)
val is_empty : open_env -> bool

(**
*@param old_oe => OpenEnv.open_env
*@param invocation_map => JBasics.class_method_signature list JBasics.ClassMethodMap.t
*@return bool
*)
val check_rm_anns : open_env -> (JBasics.class_method_signature list) ClassMethodMap.t -> bool
